# downloader.py

import multiprocessing
import functools
import logging

from numpy import ndarray
from pandas import Series

import py621dl.reader
from py621dl.net import get_image


class E621Downloader:

    def __init__(self, csv_reader: py621dl.reader.Reader, /, *, timeout=5, retries=3):
        self.reader = csv_reader
        self.batch_size = self.reader.batch_size
        self.timeout = timeout
        self.retries = retries

    def __iter__(self):
        return self

    def __next__(self) -> tuple[list[ndarray], list[Series]]:
        images = []
        rows = []
        try:
            while len(images) < self.batch_size:
                if images:
                    logging.info(f"Retrying {self.batch_size - len(images)} images")
                    batch = self.reader.get_rows(self.batch_size - len(images))
                else:
                    batch = next(self.reader)

                if self.batch_size > 2:
                    with multiprocessing.Pool(processes=min(multiprocessing.cpu_count(), self.batch_size)) as pool:
                        images_temp = pool.map_async(
                            functools.partial(get_image, timeout=self.timeout, retries=self.retries),
                            [row['md5'] for row in batch]
                        )

                        images_temp.wait()
                        images_temp = images_temp.get()
                else:
                    images_temp = [get_image(row['md5'], timeout=self.timeout, retries=self.retries) for row in batch]

                exc = [e for e in images_temp if isinstance(e, Exception)]
                img = {i: img for i, img in enumerate(images_temp) if isinstance(img, ndarray)}

                if exc:
                    exc = ExceptionGroup("One or more downloads failed", exc)
                    logging.error(f'Failed to download images: {exc}\nAttempting to fill batch')

                images.extend(img.values())
                rows.extend([batch[i] for i in img.keys()])

            return images, rows
        except StopIteration:
            raise StopIteration("No more rows to read")



# reader.py

import json
from os.path import exists

import pandas as pd


DATA_COLUMNS = ["id", "md5", "rating", "image_width", "image_height", "tag_string", "score", "is_deleted", "is_flagged"]


class Reader:

    def __init__(self, csv_file, /, batch_size=2, excluded_tags=None, minimum_score=10, *, chunk_size=2000,
                 checkpoint_file=None, repeat=True):
        self.__csv = csv_file
        self.__checkpoint_file = checkpoint_file
        self.__repeat = repeat

        self.__excluded_tags = frozenset(excluded_tags) if excluded_tags is not None else None
        self.__minimum_score = minimum_score

        self.__df_buffered = pd.read_csv(csv_file, chunksize=chunk_size, usecols=DATA_COLUMNS, iterator=True)
        self.__chunk_iter = None
        self.__row_iter = None

        self.__chunk_idx = None
        self.__row_idx = None

        self.batch_size = batch_size

        self.__init()

    def get_rows(self, n_rows=1):
        batch = []
        while len(batch) < n_rows:
            try:
                self.__row_idx, row = next(self.__row_iter)
                batch.append(row)
            except StopIteration:
                try:
                    self.__chunk_idx, chunk = next(self.__chunk_iter)
                except StopIteration:
                    if self.__repeat:
                        self.__df_buffered = pd.read_csv(self.__csv, chunksize=2000,
                                                         usecols=DATA_COLUMNS, iterator=True)
                        if self.__checkpoint_file is not None:
                            with open(self.__checkpoint_file, 'w') as f:
                                json.dump([0, 0], f)
                        self.__init(repeat_init=True)
                        continue
                    else:
                        raise StopIteration
                else:
                    self.__row_iter = iter(self.__row_gen(chunk))
                    continue
        self.__save_checkpoint()
        return batch

    def get_row_idx(self):
        return self.__row_idx

    def get_chunk_idx(self):
        return self.__chunk_idx

    def __init(self, repeat_init=False):
        self.__chunk_iter = iter(self.__chunk_gen())
        # check if checkpoint is None, if not try loading it as json
        if self.__checkpoint_file is not None and exists(self.__checkpoint_file) and not repeat_init:
            with open(self.__checkpoint_file, 'r') as f:
                chunk_idx_skip, row_idx_skip = json.load(f)

            self.__chunk_idx, chunk = next(self.__chunk_iter)
            self.__row_idx = 0

            # skip chunks
            while self.__chunk_idx < chunk_idx_skip:
                self.__chunk_idx, chunk = next(self.__chunk_iter)

            self.__row_iter = iter(self.__row_gen(chunk))

            # skip rows
            while self.__row_idx != row_idx_skip:
                self.__row_idx, _ = next(self.__row_iter)

        else:
            self.__chunk_idx, self.chunk = next(self.__chunk_iter)
            self.__row_iter = iter(self.__row_gen(self.chunk))

    def __chunk_gen(self):
        for chunk_idx, chunk in enumerate(self.__df_buffered):
            yield chunk_idx, chunk

    def __row_gen(self, chunk):
        for row_idx, row in chunk.iterrows():

            if self.__is_row_invalid(row):
                continue

            yield row_idx, row

    def __save_checkpoint(self):
        if self.__checkpoint_file is not None:
            with open(self.__checkpoint_file, 'w') as f:
                json.dump([self.__chunk_idx, self.__row_idx], f)

    def __is_row_invalid(self, row):
        is_invalid = False
        is_invalid |= bool(self.__excluded_tags) and self.__excluded_tags.intersection(row['tag_string'].split()) != set()
        is_invalid |= self.__minimum_score > row['score']
        is_invalid |= row['is_deleted'] == 't'
        is_invalid |= row['is_flagged'] == 't'
        return is_invalid

    def __iter__(self):
        return self

    def __next__(self):
        try:
            return self.get_rows(self.batch_size)
        except StopIteration:
            raise StopIteration



# net.py

import io
import logging
import os

import cv2
import numpy as np
import requests
from PIL import Image, UnidentifiedImageError
from requests.exceptions import HTTPError, ConnectionError, Timeout

from py621dl.exceptions import CVException


# noinspection PyTypeChecker, PyUnresolvedReferences
def get_image(md5, /, timeout=5, retries=3, *, __is_test_target=False):
    url = construct_e621_img_link(md5)

    if url == "test":
        return np.zeros((100, 100, 3), dtype=np.uint8)  # for testing

    download = download_image_as_bytes(url, timeout, retries)

    if isinstance(download, Exception):
        raise download

    try:
        img = io.BytesIO(download)
    except OSError as e:
        logging.error(f'Failed to convert bytes to image: {e}')
        return e

    try:
        img = Image.open(img)
    except UnidentifiedImageError as e:
        logging.error(f'Failed to identify image format from url: {url}')
        return e

    try:
        img = cv2.cvtColor(np.array(img), cv2.COLOR_RGB2BGR)
    except CVException as e:
        logging.error(f'Image decode failed: {e}')
        return e

    return img


def download_image_as_bytes(url, timeout=5, retries=3):
    exc = None
    for i in range(retries):
        try:
            response = requests.get(url, timeout=timeout)
            response.raise_for_status()
        except HTTPError as e:
            logging.error(f'HTTP Error: {e}')
            return e
        except (ConnectionError, Timeout) as e:
            logging.error(f'Connection or Timeout Error: {e}\nRetrying attempt {i + 1} of {retries}')
            return e
        else:
            return requests.get(url).content
    else:
        logging.error(f'Failed to download image after {retries} attempts')
        return exc


def construct_e621_img_link(post_md5):
    if post_md5 == "test":
        return post_md5

    return f'https://static1.e621.net/data/{post_md5[0:2]}/{post_md5[2:4]}/{post_md5}.jpg'



# exceptions.py

class CVException(Exception):
    """Raised for color conversion and image decoding error from downloaded image."""

    def __init__(self):
        super().__init__("Failed to convert image to OpenCV format or convert color space")



# __init__.py

import logging
import tomllib
from importlib import resources

from py621dl.downloader import E621Downloader
from py621dl.reader import Reader


def enable_logging(level: int = logging.INFO, /, *, filename: str = None):
    logger = logging.getLogger()
    logger.disabled = False

    if filename:
        handler = logging.FileHandler(filename)
    else:
        handler = logging.StreamHandler()

    handler.setFormatter(logging.Formatter('%(asctime)s - %(levelname)s - %(message)s'))
    handler.setLevel(level)
    logger.addHandler(handler)


logging.getLogger().disabled = True

__version__ = "0.1a0.dev2"
resource_folder = resources.files("py621dl")

_cfg = tomllib.loads(resource_folder.joinpath("config.toml").read_text())



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['py621dl']

package_data = \
{'': ['*']}

install_requires = \
['Pillow', 'numpy', 'opencv-python', 'pandas', 'requests']

setup_kwargs = {
    'name': 'py621dl',
    'version': '0.1a0.dev3',
    'description': 'A downloader for e621.net with meant for iterability, using e621 db post exports, designed for deep learning.',
    'long_description': '# py621dl - an iterable E621 downloader\n\nThis package is meant to be used in deep learning applications and automation,\nnot as a means to download specific images and post IDs or searching for tags.\nFor that application, please check out [py621](https://pypi.org/project/py621/)\nwhich is not related to this package in any way.\n\nThe package is meant to be used with the official db export format from E621,\nposts information. See [here](https://e621.net/db_export/) for available db exports\nand [here](https://e621.net/help/api) for general information on the API.\n\n!! This is a pre-release version, and is not meant for production use !!\n\nProper documentation, tests, and automated updates to the package will be added later.\n\n## Usage\n\nThe E621Downloader class must be initialized using the Reader class, to which\nthe csv file must be passed. The Reader supports only the official db export csv files\nof the format "posts-YYYY-MM-DD.csv.gz", either compressed or uncompressed.\n\nThe [E621Downloader](https://github.com/slobodaapl/py621dl/blob/90228111c166926f63f0cf3b991d9c82ca79e6e8/src/py621dl/downloader.py#L12-L14)\nclass can be initialized with the following parameters:\n\n- `csv_reader`:\n  the [Reader](https://github.com/slobodaapl/py621dl/blob/90228111c166926f63f0cf3b991d9c82ca79e6e8/src/py621dl/reader.py#L10-L13)\n  object\n- `timeout`: the timeout for the requests, in seconds\n- `retries`: the number of retries for the requests\n\nIt can be used as an iterable, yielding lists of `np.ndarray` objects of the images. The list size\nwill depend on your `batch_size` specified for `Reader`. The images are of opencv BGR format.\nThe downloader automatically handles and filters deleted or flagged posts, and will attempt to fill\nthe batch with new images so that it will always yield a full batch.\n\nThe [Reader](https://github.com/slobodaapl/py621dl/blob/90228111c166926f63f0cf3b991d9c82ca79e6e8/src/py621dl/reader.py#L10-L13)\nclass can be initialized with the following parameters:\n\n- `csv_file`: the path to the csv file\n- `batch_size`: the size of the batch to be returned by the `E621Downloader`\n- `excluded_tags`: a list of E621 tags to be excluded from the results\n- `minimum_score`: the minimum score of the posts to be included in the results\n- `chunk_size`: the size of the chunk to be read from the csv file at once\n- `checkpoint_file`: the path to the checkpoint file, to resume from any point. If path doesn\'t exist, a new file will\n  be created.\n- `repeat`: whether to repeat from the beginning of the csv file when the end is reached automatically.\n  Otherwise `StopIteration` is raised.\n  `E621Downloader` handles this exception and raises its own `StopIteration` when the end is reached.\n\n# Example use\n\n```python\nfrom py621dl import Reader, E621Downloader\n\nreader = Reader("posts-2022-10-30.csv.gz")\ndownloader = E621Downloader(reader, timeout=10, retries=3)\n\nfor batch in downloader:\n    # do something with the batch\n    pass\n```',
    'author': 'slobodaapl',
    'author_email': 'slobodaapl@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://pypi.org/project/py621dl',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.11',
}


setup(**setup_kwargs)
